# `go_sim_lib.py` is a miniature library to run Cytosim.
#  It is not executable by directly, but it used by go_sim.py
#  to create directory, copy files, move directories, etc.
#
# Copyright F. Nedelec 2007--2019, S. Dmitrieff 2019


try:
    import os, shutil, subprocess
except ImportError:
    import sys
    host = os.getenv('HOSTNAME', 'unknown')
    sys.stderr.write("go_sim_lib.py could not load python modules on %s\n" % host)
    sys.exit()

try:
    import exceptions
except:
    try:
        import builtins as exceptions
    except:
        import sys
        host = os.getenv('HOSTNAME', 'unknown')
        sys.stderr.write("go_sim_lib.py could not load `exceptions` on %s\n" % host)
        sys.exit()

class Error( exceptions.Exception ):
    """go_sim.py exception class"""
    def __init__(self, value=None):
        self.value = value
    def __str__(self):
        return repr(self.value)

# default name of the config file:
config_name = 'config.cym'
logfile_name = 'log.txt'


#==========================  DIR/FILES HANDLING ==============================

def make_temp_directory():
    #make a directory on the /scratch disc if possible:
    str = os.getenv('USER', 'run') + '-'
    import tempfile
    try:
        return tempfile.mkdtemp('', str, '/scratch/ned')
    except:
        pass
    try:
        return tempfile.mkdtemp('', str, '/scratch')
    except:
        pass
    tmp = os.getenv('TMPDIR', '')
    if tmp:
        return tmp
    return tempfile.mkdtemp('', str, '.')


def make_directory(path, n=0):
    """
    Create a new directory name????,
    where ???? is a 4-digit number greater than n
    """
    res = path
    if path[-1].isdigit():
        path = path + '-'
    while n < 10000:
        try:
            os.mkdir(res)
            #print("made " + res)
            return res
        except OSError:
            res = path + '%04i' % n
        n += 1
    raise Error("failed to create new run directory on "+os.getenv('HOSTNAME', 'unknown'))


def copy_recursive(src, dst):
    """recursively copy everything from src to dst"""
    if os.path.isfile(src):
        shutil.copy2(src, dst)
    elif os.path.isdir(src):
        try:
            os.mkdir(dst)
        except OSError:
            pass
        files = os.listdir(src)
        for f in files:
            s = os.path.join(src, f)
            d = os.path.join(dst, f)
            copy_recursive(s, d)


def move_directory(path, park, name):
    """Copy directory 'path' to park, under a similar name"""
    src = os.path.abspath(path)
    if src == os.path.abspath(park):
        return src
    try:
        dst = make_directory(os.path.join(park,name))
    except:
        sys.stderr.write("go_sim_lib.py found no parking space for '%s'\n" % path)
        return src
    #print("moving directory( %s -> %s )" % (src, dst))
    copy_recursive(src, dst)
    from filecmp import dircmp
    dcmp = dircmp(src, dst)
    if dcmp.left_only or dcmp.diff_files:
        sys.stderr.write("go_sim_lib.py could not copy '%s' identically\n" % path)
        return src
    else:
        shutil.rmtree(src)
    return dst


def copy_config(conf, repeat):
    """
        make 'repeat' copies of the config files.
    """
    res = []
    for x in range(repeat):
        res.extend([conf]);
    return res


def make_config(conf, repeat, preconf, dest):
    """
    Generate config files by running a preconf script,
    or simply repeat the name if ( repeat > 1 ) and preconf==''.
    """
    if preconf:
        module = {}
        try:
            module = __import__(preconf.rstrip('.py'))
        except:
            import imp
            module = imp.load_source('pre_config', preconf)
        if not module:
            raise Error("could not load python module `"+preconf+"'")
        # use module to generate a new config file:
        return module.parse(conf, {}, repeat, dest)
    else:
        return copy_config(conf, repeat)


#=======================  RUNNING THE SIMULATION  ==============================

def run_sim(exe, args):
    """
    Start executable in current directory, and wait for completion.
    Standard output is sent to `out.txt' and standard error to `err.txt'.
    The executable shall find its default configuration file.
    """
    outname = 'out.txt'
    errname = 'err.txt'
    outfile = open(outname, 'w')
    errfile = open(errname, 'w')
    # run simulation
    if not args:
        val = subprocess.call(exe, stdout=outfile, stderr=errfile)
    else:
        val = subprocess.call([exe]+args, stdout=outfile, stderr=errfile)
    outfile.close()
    errfile.close()
    # remove output files if empty:
    if os.path.isfile(outname) and not os.path.getsize(outname):
        os.remove(outname)
    if os.path.isfile(errname) and not os.path.getsize(errname):
        os.remove(errname)
    return val


def info_start(filename, exe, args, conf, pid):
    import time
    with open(filename, "w") as f:
        f.write("host      %s\n" % os.getenv('HOSTNAME', 'unknown'))
        f.write("user      %s\n" % os.getenv('USER', 'unknown'))
        f.write("wdir      %s\n" % os.getcwd())
        f.write("exec      %s\n" % exe)
        f.write("args      %s\n" % args)
        f.write("conf      %s\n" % conf)
        f.write("pid       %s\n" % pid)
        f.write("start     %s\n" % time.asctime())


def info_end(filename, val):
    import time
    with open(filename, "a") as f:
        f.write("status    %s\n" % val)
        f.write("stop      %s\n" % time.asctime())


def run(exe, conf, name, args=[]):
    """
    Run one simulation in a new sub directory and wait for completion.
    The config file 'conf' is copied to the subdirectory.
    Returns sub-directory in which `exe` was called.
    """
    cdir = os.getcwd()
    if not os.path.isfile(conf):
        raise Error("missing/unreadable config file")
    conf = os.path.abspath(conf);    
    # use a temporary directory on the cluster:
    if 'SLURM_JOB_ID' in os.environ or 'LSB_JOBID' in os.environ:
        wdir = make_temp_directory()
    else:
        wdir = make_directory(name)
    os.chmod(wdir, 504)
    os.chdir(wdir)
    shutil.copyfile(conf, config_name)
    info_start(logfile_name, exe, args, conf, os.getpid())
    val = run_sim(exe, args)
    info_end(logfile_name, val)
    os.chdir(cdir)
    return (val, wdir)


def start(exe, conf, name, args=[]):
    """
    Start simulation in a new sub directory, and return immediately.
    The config file `conf` is copied to the sub-directory.
    """
    cdir = os.getcwd()
    if not os.path.isfile(conf):
        raise Error("missing/unreadable config file")
    conf = os.path.abspath(conf);
    wdir = make_directory(name)
    os.chdir(wdir)
    shutil.copyfile(conf, config_name)
    if exe != 'none':
        outfile = open('out.txt', 'w')
        errfile = open('err.txt', 'w')
        #start simulation, but do not wait for completion:
        pid = subprocess.Popen(['nohup', exe]+args, stdout=outfile, stderr=errfile).pid
        info_start(logfile_name, exe, args, conf, pid)
    else:
        pid = 0
    os.chdir(cdir)
    return (pid, wdir)





#!/usr/bin/env python3
#
# get_data.py
#
# simple extraction of data from files
#
# Copyright F. Nedelec, 2011 - 2016

"""
    Collect data from given run directories, and print them to standard output

Syntax:

    get_data.py DIRECTORIES
    
Example:

    get_data run???? > data.txt

Description:

    This script must be customized for any meaningful application,
    but it should be a useful template to start from.
    Please make a copy of the script with a different name.

F. Nedelec, Jan. 2016
"""

import sys, os, subprocess


#------------------------------------------------------------------------

def find_value(file, key):
    """
        Find values corresponding to 'key' in the file.
        Multiple values are concatenated.
    """
    res = ''
    for line in file:
        inx = line.find(key)
        if inx >= 0:
            r = line.find('=', inx)
            if r > inx:
                v = line[r+1:].strip()
                if len(res):
                    res += '|' + v
                else:
                    res = v
    return res


def get_cell(file, ii, jj):
    """
        Extract the jj-th word from line ii
        If jj is an array, the corresponding values will be concatenated
    """
    if len(jj) == 0:
        return '';
    cnt = 0
    for line in file:
        cnt = cnt + 1
        if cnt == ii:
            s = line.split()
            val = s[jj[0]]
            for n in jj[1:]:
                val = s[n] + ' '
            return val
    return ''


def get_column(file, jj):
    """
        Extract the jj-th word from line ii
        If jj is an array, the corresponding values will be concatenated
        """
    res = ' '
    for line in file:
        s = line.split()
        if not s or s[0] == '%':
            continue
        try:
            res += s[jj].rjust(7) + ' '
        except IndexError:
            pass
    return res


def find_differences(left, right):
    """
        Find differences between two files that are numeric values
        Multiple values are concatenated.
        This can be used to compare config files
        """
    if not os.path.isfile(left):
        return 'file not found: ' + left
    if not os.path.isfile(right):
        return 'file not found: ' + right
    res = ''
    sub = subprocess.Popen(['diff', left, right], stdout=subprocess.PIPE)
    for stuff in sub.stdout:
        try:
            line = stuff.decode('utf-8')
        except:
            line = stuff
        if line[0] == '>':
            for s in line.split():
                s = s.rstrip(',;')
                try:
                    int(s)
                except:
                    try:
                        float(s)
                    except:
                        continue
                res += ' %-10s' % s
    sub.stdout.close()
    return res


#------------------------------------------------------------------------

def process(path):
    """ 
        This extracts parameters from the config file,
        and values from 'mom.txt'
    """
    res = path + ' '
    try:
        res += find_differences('config.cym', path+'/config.cym')
    except IOError as e:
        sys.stderr.write("Error: %s\n" % repr(e))
        return ''
    res += ' nan'
    try:
        with open(path+'/mom.txt', 'r') as f:
            res += get_column(f, -1)
    except IOError as e:
        sys.stderr.write("Error: %s\n" % repr(e))
        res = ''
    return res


def main(args):
    paths = []
    
    for arg in args:
        if os.path.isdir(arg):
            paths.append(arg)
        else:
            sys.stderr.write("  Error: unexpected argument `%s'\n" % arg)
            sys.exit()
    
    if not paths:
        sys.stderr.write("  Error: you must specify directories\n")
        sys.exit()
    
    nb_columns = 0
    for p in paths:
        res = process(p)
        # check that the number of column has not changed:
        if nb_columns != len(res.split()):
            if nb_columns == 0:
                nb_columns = len(res.split())
            else:
                sys.stderr.write("Error: data size mismatch in %s\n" % p)
                return
        print(res)


#------------------------------------------------------------------------

if __name__ == "__main__":
    if len(sys.argv) < 2 or sys.argv[1].endswith("help"):
        print(__doc__)
    else:
        main(sys.argv[1:])





#!/usr/bin/env python3
#
# tell.py prints parameters from config files in given directories
#
# F. Nedelec, Jan 24th, 2008, March 2010, July 2014, July 2017

"""
    Print parameters from multiple files in CSV column format
    
Syntax:
    
    tell.py parameter-names [directories/files]

"""


import sys, os
sep = ' '


def format_value(val):
    if isinstance(val, str):
        return val
    if isinstance(val, dict):
        r = ""
        s = "{ "
        for k in sorted(val):
            r += s + str(k) + " = " + format_value(val[k])
            s = "; "
        return r + "; }"
    try:
        r = ""
        s = ""
        for k in val:
            r += s + format_value(k)
            s = ", "
        return r
    except:
        return str(val)


def column_width(table):
    """calculate the maximum width of each column"""
    w = []
    for line in table:
        for i, l in enumerate(line):
            while i >= len(w):
                w.append(0)
            v = len(format_value(l))
            if v > w[i]:
                w[i] = v
    return w


def format_table(table):
    """print table in tidy column format"""
    width = column_width(table)
    indx = range(1,len(width))
    res = ""
    for line in table:
        val = format_value(line[0]).ljust(width[0])
        for i in indx:
            if not line[i]:
                v = "-"
            else:
                v = format_value(line[i])
            val += sep+v.rjust(width[i])
        res += val + '\n'
    return res


def config_file(arg):
    """return name of probable config-file"""
    if os.path.isdir(arg):
        file = os.path.join(arg, 'config.cym')
        if os.path.isfile(file):
            return file
    if os.path.isfile(arg):
        return arg
    return ""


def find_any(str, chars):
    r = len(str)
    for c in chars:
        x = str.find(c)
        if x >= 0 and x < r:
            r = x
    return r


def find_value(file, key):
    """find value corresponding to key. Multiple values are concatenated"""
    val = []
    for line in file:
        [k, e, values] = line.partition('=')
        if key == k.strip() and e == '=':
            #print(k, e, values)
            e = find_any(values, '\n%')
            val.append(values[0:e].strip())
    return val


def tell(files, keys):
    """print parameter from multiple files in columns"""
    res=[]
    line = ['% file']
    line.extend(keys)
    res.append(line)
    
    for f in files:
        try:
            file = open(f, "r")
            vals = []
            for key in keys:
                file.seek(0)
                vals.append(find_value(file, key))
            file.close()
            if f.endswith('/config.cym'):
                line = [f[:len(f)-11]]
            else:
                line = [f]
            line.extend(vals)
            res.append(line)
        except IOError:
            continue
    return res;


#------------------------------------------------------------------------

def main(args):
    """print parameter from multiple files in columns"""
    global sep
    paths = []
    keys = []
    for arg in args:
        if os.path.exists(arg):
            paths.append(arg)
        elif arg.startswith('sep='):
            sep = arg[4:]
        else:
            keys.append(arg)
    
    if not paths:
        print(__doc__)
        return

    files = []
    for p in paths:
        files.append(config_file(p))
    
    table = tell(files, keys)
    #table = [ ['a', '1'], ['b', '2'], ['c', '3'] ]
    print(format_table(table))


if __name__ == "__main__":
    if len(sys.argv)>1 and sys.argv[1].endswith("help"):
        print(__doc__)
    else:
        main(sys.argv[1:])




#!/usr/bin/env python
# A script to run simulations sequentially.
# Copyright F. Nedelec, 2010--2018
# Using multiprocessing thanks to Adolfo Alsina and Serge Dmitrieff, March 2016

"""
Synopsis:
    
    Run simulations sequentially.
    For each config file, a simulation is started in a separate 'run' directory.
    Completed runs are moved to the 'park' directory if specified.

Syntax:

    go_sim.py [executable] [repeat] [script=PYTHON] [park=directory] config_file [config_file]
    
    Bracketted arguments are optional.
    If working_directory is not specified, the current directory is used.
    [repeat] is an integer specifying the number of run for each config file.
    Completed simulations will be store in the 'park' directory if specified.
    
    If a python script is specified, it should provide a function parse(input)
       You may use: preconfig.py
    
    Any number of config file can be provided (at least one).

F. Nedelec, 03.2010, 10.2011, 05.2012, 04.2013, 12.2017
"""

# Loading modules on the compute-cluster may fail for unexpected reasons.
# The long time.sleep() prevents zoombie nodes from accepting further LSF jobs

try:
    import os, sys, time
except ImportError:
    host = os.getenv('HOSTNAME', 'unknown')
    sys.stderr.write("go_sim.py could not load necessary python modules on %s\n" % host)
    time.sleep(10000)
    sys.exit()

# go_sim.py ignores interupting SIGNALS, to make sure that it can perform cleaning-up operations
# if the child executable is killed on the same occasion.
# This way 'CTRL-C' will kill the executable, but not this controlling script.

def handle_signal(sig, frame):
    sys.stderr.write("go_sim.py escaped signal %i\n" % sig)


try:
    import signal
    signal.signal(signal.SIGINT, handle_signal)
    signal.signal(signal.SIGTERM, handle_signal)
    #sys.stderr.write("go_sim.py registered its signal handler\n")
except ImportError:
    host = os.getenv('HOSTNAME', 'unknown')
    sys.stderr.write("go_sim.py could not load `signal` on %s\n" % host)
    pass


#define output for error messages:
err    = sys.stderr
out    = sys.stdout

njobs  = 1
repeat = 1
park   = ''
exe    = os.path.abspath('sim')

try:
    import go_sim_lib
except ImportError:
    err.write("go_sim.py could not load go_sim_lib.py\n")
    sys.exit()


#------------------------------------------------------------------------

def executable(arg):
    return os.path.isfile(arg) and os.access(arg, os.X_OK)


def run(conf, name):
    """
        run executable 'exe' with config 'conf' in a directory of name 'name'
    """
    try:
        (val, res) = go_sim_lib.run(exe, conf, name)
        if val == 0:
            out.write("Completed run `%s` in %s\n" % (conf, res))
        else:
            out.write("Failed run `%s` in %s with value %i\n" % (conf, res, val))
    except KeyboardInterrupt:
        err.write("go_sim.py `%s` was interrupted\n" % conf)
    # move run directory to `park` if specified:
    if os.path.isdir(park):
        try:
            res = go_sim_lib.move_directory(res, park, name)
            with open(res+"/log.txt", "a") as f:
                f.write("parked    %s\n" % time.asctime())
            out.write("            ---> parked in %s\n" % res)
        except Exception as e:
            err.write("go_sim.py cannot move directory: %s\n" % repr(e))


def run_queue(queue):
    """
    run items in the queue
    """
    while True:
        try:
            arg = queue.get(True, 1)
            run(*arg)
        except:
            break;


def process(conf, preconf, name, queue):
    """
        run configurations files generated from 'conf'
    """
    if not os.path.isfile(conf):
        err.write("go_sim.py: file '%s' does not exist\n" % conf)
        sys.exit()
    
    # generate config file(s):
    if preconf:
        import tempfile
        tmp = tempfile.mkdtemp('', name+'-', '.')
        files = go_sim_lib.make_config(conf, repeat, preconf, tmp)
    else:
        files = go_sim_lib.copy_config(conf, repeat)

    if not files:
        err.write("go_sim.py could not generate config files\n")
        sys.exit()
 
    # process all files created:
    for i, f in enumerate(files):
        if len(files) > 1:
            n = name + '-%04i' % i
        else:
            n = name
        if njobs > 1:
            queue.put((f, n))
            #print('Queued ' + f + ' ' + n)
        else:
            run(f, n)


#------------------------------------------------------------------------

def main(args):
    global njobs, repeat, park, exe, queue
    preconf = ''
    name    = 'run0000'
    files   = []
    njobs   = 1
    
    # parse arguments list:
    for arg in args:
        if arg.isdigit():
            repeat = int(arg)
        elif arg.startswith('nproc=') or arg.startswith('njobs='):
            njobs = int(arg[6:])
        elif arg.startswith('jobs='):
            njobs = int(arg[5:])
        elif arg.startswith('script='):
            preconf = arg[7:]
        elif executable(arg):
            exe = os.path.abspath(arg)
        elif os.path.isfile(arg):
            files.append(arg)
        elif arg.startswith('name='):
            name = arg[5:]
        elif arg.startswith('park='):
            park = arg[5:]
            if not os.path.isdir(park):
                err.write("go_sim.py: `%s' is not a directory\n" % park)
                sys.exit()
        else:
            err.write("go_sim.py: unexpected argument `%s'\n" % arg)
            err.write("         : there is no file with that name\n")
            sys.exit()
        
    if not files:
        err.write("You should specify a config file on the command line\n")
        sys.exit()

    if not executable(exe):
        err.write("go_sim.py: executable '%s' not found\n" % exe)
        sys.exit()

    # prepare for multiprocessing
    queue = ()
    if njobs > 1:
        try:
            from multiprocessing import Process, Queue
            queue = Queue()
        except ImportError:
            out.write("Warning: multiprocessing unavailable\n")
            njobs = 1

    # process all given files:
    cnt = 0
    for conf in files:
        process(conf, preconf, name, queue)
        cnt += 1
        name = 'run%04i' % cnt

    # process jobs:
    if njobs > 1:
        jobs = [Process(target=run_queue, args=(queue,)) for n in range(njobs)]
        for job in jobs:
            job.start()
        for job in jobs:
            job.join()
    return 0


#------------------------------------------------------------------------


if __name__ == "__main__":
    if len(sys.argv) < 2 or sys.argv[1].endswith("help"):
        print(__doc__)
    else:
        main(sys.argv[1:])





#!/usr/bin/env python3
#
# PRECONFIG, a versatile configuration file generator for varying parameters
#
# Copyright Francois J. Nedelec and  Serge Dmitrieff, 
# EMBL 2010--2017, Cambridge University 2019--2022
# This is PRECONFIG version 1.59, last modified on 03.02.2023

"""
# SYNOPSIS

   Preconfig generates files from a template by evaluating doubly-bracketed Python code.
   
# Article:

   preconfig: A Versatile Configuration File Generator for Varying Parameters
   https://openresearchsoftware.metajnl.com/articles/10.5334/jors.156/

# DESCRIPTION

   Preconfig reads the template file from top to bottom, identifying snippets
   of code which are surrounded by double square brackets. It then executes this
   code using the python interpreter, proceeding recursively whenever multiple
   values are specified. Values are eventually converted to their string
   representation, and substituted in place of the code snippet. In this way,
   Preconfig will generate all the possible combinations following the order in
   which these combinations were specified in the file. Importantly, any ac-
   -companying text in the template file is copied verbatim to the output file,
   such that any syntax present in the configuration file can be maintained
   during the process.

   At least one template file should be specified; other arguments are optional.
   If several template files are specified, they will be processed sequentially.
   The names of the produced files are built from the name of the template
   by removing any second extension, and inserting an integer of constant width.

   Examples:

   - config.cym.tpl --> config0000.cym, config0001.cym, config0002.cym, etc.
   - config.txt.tpl --> config0000.txt, config0001.txt, config0002.txt, etc.
   - model.xml.tpl --> model0000.xml, model0001.xml, model0002.xml, etc.

   The width of the variable part (default=4) can be changed on the command line.
   For instance, to specify a width of 2 characters, invoke "preconfig -2 ...".

   By default, indexing starts at zero, but this can be changed with 'n=INTEGER'.

# SYNTAX

   preconfig [OPTIONS] [DEFINITIONS] TEMPLATE_FILE [ADDITIONAL_TEMPLATE_FILES]

# OPTIONS

    - a pattern can be specified as 'path=STRING', to define the names of
    the files to be generated, for example 'config%04i.cym'. If the pattern
    includes '/', as in 'run%04i/config.cym', the directories will be made.
    The pattern must accept integer substitution (eg. '%i' or '%4i' or '%04i').

   - if a positive integer REPEAT is specified, each template file will be
   processed REPEAT times, for example: `preconfig 3 config.cym.tpl` will parse
   the template three times and generate three times more files.

   - if the path to an existing directory is specified, files will be created
   in this directory, for example: `preconfig dir config.cym.tpl`
   
   - if 'path=FILENAME' is specified, Preconfig will use this name for the first
   file that it generates. If this file already exists, it will be overwritten.

   - if a negative integer is specified, this will set the width of the integer
   that is used to build the file namess.
   For example: `preconfig -2 config.cym.tpl` will create 'config00.cym', etc.

   - if a '-' is specified, all accessory output is suppressed

   - if a '+' is specified, more detailed information on the parsing is provided.

   - if '++' or 'log' is specified, a file 'log.csv' will be created containing one
   line for each file created, containing the substitutions operated for this file.

   - if '--help' is specified, this documentation will be printed.
   
# DEFINITIONS

   Variables can be defined on the command line as 'name=value', with no space 
   around the '='. They are added to the dictionary used to evaluate the code 
   snippets found inside the template file.
   example: `preconfig rate=100 config.cym.tpl`

   Sequences can be defined as 'name=sequence' or 'name==sequence'. In the first
   instance, the sequence is expanded, generating new files for each value.
   However, using '==' prevents this expansion, and the variable is used verbatim.
   example: `preconfig rate=[1,10,100] config.cym.tpl`

# CODE SNIPPETS

   Any plain python code can be embedded in the file, and functions from the
   [Random Module](https://docs.python.org/library/random.html) can be used.
   It is possible to use multiple bracketed expressions in the same file, and
   to define variables in the python environment.
   The index of the file being generated is defined as variable 'n'.

   Note that variables defined in embedded code are not expanded when they appear
   alone in another code (eg. [[vec]]). Furthermore, Preconfig will keep any code
   that it cannot evaluate verbatim, which happens for example if they contain
   a variable that is not defined, or some syntax error.

## Example 1

   Generate all combinations with multiple values for 2 parameters:

    rate = [[ [1, 10, 100] ]]
    speed = [[ [-1, 0, 1] ]]

   Command: `preconfig TEMPLATE_FILE` with the appropriate file name.
   In this case, Preconfig will generate 9 files.

## Example 2

   Regularly scan 2 parameters with 10 values each,
   one according to a linear scale, and the other with a geometric scale:

    [[ x = range(10) ]]
    [[ y = range(10) ]]
    reaction_rate = [[ 1 + 0.5 * x ]]
    diffusion_rate = [[ 1.0 / 2**y ]]

   Command: `preconfig TEMPLATE_FILE`
   In this case, Preconfig will generate 100 files, one for each combination.

## Example 3

   Scan multiple parameters values randomly:

    diffusion_rate = [[ random.uniform(0,1) ]]
    binding_rate = [[ round(random.uniform(0,1), 3) ]]
    reaction_rate = [[ random.choice([1, 10, 100]) ]]
    abundance = [[ random.randint(0, 1000) ]]

   Command: `preconfig 256 TEMPLATE_FILE`
   In this case, Preconfig is instructed to generate 256 files.

## Example 4

   Randomize two parameters while keeping their ratio constant:

    [[ x = random.uniform(0,1) ]]
    binding_rate = [[ 10.0 * x ]]
    unbinding_rate = [[ x ]]

   Command: `preconfig 256 TEMPLATE_FILE` to make 256 files.

## Example 5

   Randomize one parameter, using 256 values in ascending order:

    [[ x = sorted([random.uniform(0.10, 0.25) for i in range(256)]) ]]
    binding_rate = [[ x ]]

   Command: `preconfig TEMPLATE_FILE`
   In this case, the number of files (256) is specified in the template

## Example 6

   Boolean variables can be used to introduce qualitative differences:

    [[ enable = random.choice([0, 1]) ]]
    feeback = [[ random.uniform(0, 1) if (enable) else 0  ]]

   Command: `preconfig 256 TEMPLATE_FILE` to make 256 files.

## Example 7

   Randomize a value, and print this value as a comment in the file.
   The second line below the [[...]] prints a comment, from which the value
   of 'x' can be read. This can be useful to process the results later.

    [[ x = random.uniform(0,1) ]]
    %config.x= [[ x ]]
    binding_rate = [[ 10*x ]]
    unbinding_rate = [[ 2*x ]]

   Command: `preconfig 256 TEMPLATE_FILE` to make 256 files.
   To get values: awk '/%config./{sub("%config.","");print}' config0000.cym

## Example 8

   Quotations can be used to aggregate values:

    [[ vec = ['-1 0 1', '0 1 1', '-1 0 -1'] ]]
    new microtubule
    {
        position = [[vec]]
    }

## Example 9

   Cartesian sampling with filtering:

    [[ (x,y) = [ (x,y) for x in range(10) for y in range(10) if x>y ] ]]

    new particle
    {
        position = [[x]] [[y]]
    }

## Acknowledgments:

We wish to thank the members of the Nedelec group, and all users of Cytosim
for their feedback which has contributed greatly to this development.
We thanks Shaun Jackman and Steven Andrews for valuable feedback!

Copyright Francois J. Nedelec and Serge Dmitrieff
EMBL 2010--2017; Cambridge University 2019--2022
This is Free Software with no WARANTY, just hoping to be useful.
Preconfig is distributed under GPL3.0 Licence (see LICENCE)
"""

import sys

try:
    import os, io, re, time
    GLOBALS = { 'random': __import__('random'), 'math': __import__('math') }
except ImportError as e:
    sys.stderr.write("Preconfig could not load `%s`\n"%str(e))
    sys.exit(7)

#-------------------------------------------------------------------------------

__VERSION__="1.58"

__DATE__ ="29.09.2022"

# code snippets are surrounded by double square brackets:
CODE = '['
DECO = ']'


# A function to return version to be able to pip package it
def version():
    return __VERSION__


#-------------------------------------------------------------------------------

def cannonical_pattern(arg):
    """check for repeated '%' character, replacing printf syntax: %04i """
    c = arg.count('%')
    for n in range(c,1,-1):
        if arg.find('%'*n) > 0:
            return arg.replace('%'*n, '%0'+str(n)+'i', 1)
    return arg


def pop_sequence(dic, protected):
    """
        Remove an entry in the dictionary that has multiple values
    """
    for k, v in dic.items():
        if not isinstance(v, str):
            try:
                if len(v) > 1 and not k in protected:
                    dic.pop(k)
                    #print(" pop %s : %s" %(k, str(v)))
                    return (k, v)
            except:
                pass
    return ('', [])


def get_block(file, S, E):
    """
    Extract from `file` the next block starting with DOUBLE delimiters 'SS'
        and ending with matching 'EE'.
    Returns 3 values:
        - the text found before the block start
        - the block with its delimiters
        - a boolean EOF indicator
    """
    ch = file.read(1)
    sec = 0
    pre = ''
    blk = ''
    lev = 0
    while file and ch:
        pc = ch
        ch = file.read(1)
        if ch == S:
            if sec:
                lev += 1
            if pc == S:
                sec += 1
        if sec:
            blk += pc
            #print("%c%c >  lev %i sec %i" %(pc, ch, lev, sec))
        else:
            pre += pc
        if ch == E:
            if sec:
                lev -= 1
            if pc == E:
                #print("%c%c EE lev %i sec %i" %(pc, ch, lev, sec))
                if lev == -2:
                    return (pre, blk+ch, False)
    # reaching end of file:
    return (pre, blk, True)


#-------------------------------------------------------------------------------

class Preconfig:
    """ A class container for preconfig """
    def __init__(self):
        self.verbose = 2
        # local dictionary with index of file being generated
        self.locals = { 'n' : 0 }
        # variables for which expension is disabled:
        self.protected = []
        # streams for output (all output is hidden by default):
        self.out = open(os.devnull, 'w')
        self.log = []
        self.log_doc = 1
        # motif used to compose file names
        self.pattern = ''
        # number of digits used to compose `pattern`
        self.nb_digits = 4
        # list of files generated
        self.files_made = []
        # name of first output file (only one file can be made with this name)
        self.file_name = ''
        # name of current input file being processed (used for error reporting)
        self.template = ''
    
    def write_log(self, name):
        """ record some info in the log """
        keys = sorted(self.locals.keys())
        # write column headers
        if self.log_doc:
            self.log.write('%20s' % 'file')
            for k in keys:
                self.log.write(', %10s' % k)
            self.log.write('\n')
            self.log_doc = 0
        # write a line of data:
        self.log.write('%20s' % name)
        for k in keys:
            self.log.write(', %10s' % repr(self.locals[k]))
        self.log.write('\n')

    def report(self, code, key, val):
        """ print some info to output """
        if key:
            self.out.write("|%50s <-- %s\n" % (key, repr(val)) )
        else:
            self.out.write("|%50s --> %s\n" % (code, repr(val)) )

    def evaluate(self, arg):
        """ Return evaluation of `arg', checking for syntax error"""
        try:
            #print("Preconfig:evaluate " + arg)
            res = eval(arg, GLOBALS, self.locals)
        except NameError as e:
            #sys.stderr.write("NameError in `%s': %s\n" % (arg, str(e)))
            raise e
        except Exception as e:
            sys.stderr.write("\033[93m")
            sys.stderr.write("Preconfig Syntax Error in `%s': %s" % (arg, str(e)))
            sys.stderr.write("\033[0m")
            sys.stderr.write("\n")
            sys.exit(2)
        if not isinstance(res, str):
            try:
                res = list(res)
            except Exception:
                pass
        return res
    
    def try_assignment(self, code, blok):
        """
            Check if `arg` follows the format of a variable definition (X=RHS),
            and in that case return (key, self.evaluate(RHS)).
            Check for (X==RHS) with a double '==', returning (key, RHS) in that case.
            If `arg` is not an assignment, return ('', self.evaluate(arg))
        """
        k = ''
        v = code
        res = re.match(r" *([a-zA-Z]\w*) *= *(.*)", code)
        #print(" preconfig:try_assignment(%s): %s" % (code, res.groups()))
        if res and len(res.groups()) > 1:
            k = res.group(1)
            v = res.group(2).strip()
            if v[0] == '=':
                # with '==' use raw right-hand side 
                v = v[1:]
                self.protected.append(k)
                print("   protected `%s == %s'" % (k, v))
                return (k, v)
            #print("   identified `%s <- %s'" % (k, v))
        try:
            res = self.evaluate(v)
        except NameError as e:
            if self.verbose > 1:
                sys.stderr.write("\033[1m\033[96m")
                sys.stderr.write("Preconfig kept `%s' verbatim since %s" % (code, str(e)))
                sys.stderr.write("\033[0m")
                sys.stderr.write("\n")
            exit_code = 1
            #print(self.locals)
            self.report(blok, '', blok)
            return ('', blok)
        self.report(code, k, res)
        return (k, res)
    
    def process(self, file, text):
        """
            Identify and recursively substitute bracketed code blocks embedded
            in the input `file`. Generate output file evetytime EOF is reached.
        """
        output = text

        while file:
            (pre, blok, eof) = get_block(file, CODE, DECO)
            #print("%i characters +  '%s'" % (len(pre), blok))
            output += pre
            if eof:
                if blok:
                    sys.stderr.write("Preconfig Error: unclosed bracketted block in:\n")
                    sys.stderr.write("    "+blok.split('\n', 1)[0]+'\n')
                    sys.exit(2)
                # having exhausted the input, we generate a file:
                self.make_file(output)
                return
            # remove outer brackets:
            key = ''
            code = blok[2:-2].strip()
            # print("%4i characters... " % len(pre), end='')
            if code[0]==CODE and code[1]==CODE and code[-1]==DECO and code[-2]==DECO:
                # any further level of bracketting is not evaluated:
                val = code
            elif code in self.locals:
                # a defined variable is substituted but not expanded:
                val = repr(self.locals[code])
            else:
                # check the code for assigment, and evaluate code:
                key, vals = self.try_assignment(code, blok)
                # print("code block `%s' -> %s" % (blok, vals))
                if key:
                    self.locals[key] = vals
                try:
                    # use 'pop()' to probe if multiple values were specified...
                    # setting last value aside for later:
                    val = vals.pop()
                    ipos = file.tell()
                    for v in vals:
                        self.report(code, key, v)
                        if key:
                            self.locals[key] = v
                            self.process(file, output)
                        else:
                            self.process(file, output+str(v))
                        file.seek(ipos)
                    self.report(code, key, val)
                except (AttributeError, IndexError):
                    # a single value was specified:
                    val = vals
            # handle remaining value:
            if key:
                self.locals[key] = val
            else:
                output += str(val)
    
    def set_template(self, name):
        """
        Initialize variable to process template file 'name'
        """
        self.files_made = []
        self.template = name

    def set_pattern(self, name, path):
        """
        insert the number before the first '.' on the pattern
        """
        name = os.path.basename(name)
        [main, ext] = os.path.splitext(name)
        if ext == '.tpl':
            main, ext = main.split('.', 1)
            ext = '.' + ext
        self.pattern = main + '%0' + repr(self.nb_digits) + 'i' + ext
        if os.path.isdir(path):
            self.pattern = os.path.join(path, self.pattern)

    def next_file_name(self):
        """
        Generate the name of the next output file
        """
        if self.file_name:
            n = self.file_name
            self.file_name = ''
        elif self.pattern:
            n = self.pattern % self.locals['n']
        else:
            n = 'config%04i.txt' % self.locals['n']
        return n

    def make_file(self, text):
        """
        Create a file with the specified text
        """
        dst = self.next_file_name()
        #make directory if name includes a non-existent directory:
        dir = os.path.dirname(dst)
        if dir and not os.path.isdir(dir):
            os.mkdir(dir)
        with open(dst, 'w') as f:
            f.write(text)
            self.files_made.extend([dst])
        # fancy ouput:
        self.out.write(' \\'+repr(self.locals)+'\n')
        self.out.write('  \\'+('> '+dst+'\n').rjust(96, '-'))
        # write log:
        if self.log:
            self.write_log(dst)
        # get ready for next file:
        self.locals['n'] += 1

    def clear_locals(self):
        n = self.locals['n']
        self.locals.clear()
        self.locals['n'] = n

    def expand(self, values, file, text):
        """
            Calls itself recursively to remove all entries of the `values`
            dictionary that are associated with multiple values.
        """
        (key, vals) = pop_sequence(values, self.protected)
        if key:
            #print("cytosim:expand "+key+" = "+repr(vals))
            ipos = file.tell()
            for v in vals:
                values[key] = v
                self.report('', key, v)
                #print("|%50s <-- %s\n" % (key, str(v)) )
                self.expand(values, file, text)
                file.seek(ipos)
            # restore all values on upward recursion
            values[key] = vals
        else:
            #copy variables to local dictionnary:
            for k, v in values.items():
                self.locals[k] = v
            #the file index should only be copied once:
            values.pop('n', 0)
            self.process(file, text)

    def parse(self, name, file, values, repeat=1, path=''):
        """
            process one file, and return the list of files generated
        """
        self.set_template(name)
        if not self.pattern:
            self.set_pattern(name, path)
        if not os.path.isdir(path):
            self.file_name = path
        for x in range(repeat):
            try:
                self.expand(values, file, '')
            except IOError:
                sys.stderr.write("Preconfig could not load `%s`\n"%name)
                break
            self.clear_locals()
            file.seek(0)
        return self.files_made

    def main(self, args):
        """
            process arguments and perform corresponding task
        """
        repeat = 1
        inputs = []
        values = {}
        path = ''
        
        # first argument may define the pattern:
        if args[0].find('%') >= 0 and not os.path.isfile(args[0]):
            self.pattern = cannonical_pattern(args[0])
            args = args[1 :]
        
        # parse arguments:
        for arg in args:
            #print("Preconfig argument `%s'" % arg)
            if os.path.isdir(arg):
                path = arg
            elif arg.startswith("out="):
                path = arg[4:]
            elif arg.startswith("path="):
                path = arg[5:]
            elif os.path.isfile(arg):
                inputs.append(arg)
            elif arg.isdigit():
                repeat = int(arg)
            elif arg == '{{}}':
                CODE = '{'
                DECO = '}'
            elif arg == '<<>>':
                CODE = '<'
                DECO = '>'
            elif arg == '-':
                self.verbose = 0
            elif arg == '+':
                self.out = sys.stderr
                self.verbose = 0
            elif arg == '++' or arg == 'log':
                self.verbose = 2
                self.log = open('log.csv', 'w')
            elif arg.startswith('-') and arg[1:].isdigit():
                self.nb_digits = int(arg[1:])
            elif arg.find('=', 1) > 0:
                k, v = self.try_assignment(arg, 0)
                values[k] = v
            elif arg:
                sys.stderr.write("Preconfig does not understand argument `%s'\n" % arg)
                sys.exit(4)
        
        # path argument may define the pattern:
        if path.find('%') >= 0 and not os.path.isfile(path):
            self.pattern = cannonical_pattern(path)
            path = ''

        if not inputs:
            sys.stderr.write("Preconfig expects an input template file\n")
            sys.exit(3)

        for i in inputs:
            #out.write("Reading %s\n" % i)
            with open(i, 'r') as f:
                res = self.parse(i, f, values, repeat, path)
            if self.verbose > 0:
                if len(res) == 1:
                    print("generated %s" % res[0])
                else:
                    print("%i files generated: %s ... %s" % (len(res), res[0], res[-1]))


#-------------------------------------------------------------------------------

def parse(name, values, repeat=1, path=''):
    """
    Process one file, and return the list of files generated
    """
    try:
        with open(name, 'r') as file:
            return Preconfig().parse(name, file, values, repeat, path)
    except FileNotFoundError:
        print("No such file `%s`" % name)
    return []


if __name__ == "__main__":
    if len(sys.argv) < 2:
        print("You must specify a template file (for instructions, invoke with option '--help')")
    elif sys.argv[1].endswith("help"):
        print(__doc__)
    elif sys.argv[1]=='--version':
        print("This is PRECONFIG version %s (%s)" %(__VERSION__,__DATE__))
    elif sys.argv[1]=='--pipe':
        object=Preconfig()
        while 1:
            time.sleep(0.1)
            line = sys.stdin.readline()[:-1]
            if line:
                args = line.split(' ')
                object.main(args)
    else:
        object=Preconfig()
        object.main(sys.argv[1:])




#!/usr/bin/env python
#
# A script to submit jobs to the SLURM queuing system
#
# F. Nedelec, 10.2007 --- 29.09.2018
# Adapted from LSF to SLURM on 17.02.2017

"""
    Submit an array of jobs to the SLURM system, to be handled by 'go_sim.py'
    
Syntax:
    
    submit.py ARG [mem=????] [queue=????] [hours=INT] [days=INT] [cpu=INT] file1 [file2] [file3] [...]
    
    ARG is passed as arguments to 'go_sim.py' (please check go_sim.py help),
    and you must use quotes if you have multiple arguments to group them together.
    
    Unless specified otherwise, the queue is 'medium_priority'.
    The amount of requested memory (default=2GB) should be specified in MB:
       mem=1024 (for 1 GB)
       mem=512  (for 512 MB)
       ...
    
Examples:
    
    1. submit.py sim config.cym
       Submit one job to run the config file provided
    
    2. submit.py sim config1.cym config2.cym config3.cym ...
       Submit one job for each config file provided. They will run in parallel.

    3. submit.py 'sim preconfig.py' config.cym.tpl
       Submit one job where the config file will be generated by preconfig.py from config.cym.tpl
       The executable will run sequentially on one node, for each file that is generated.
    
    4. submit.py 'sim 10' config.cym
       Submit one job to run the same config file 10 times sequentially on one node.
    
    5. submit.py 'sim 10' config1.cym config2.cym config3.cym ...
       Submit a different job for each config file provided, with 10 repeats of each
    
    
Last updated 26.09.2018
F. Nedelec
"""


import sys, os, shutil, subprocess

# default parameters for submission:
subcmd  = 'sbatch'
queue   = 'htc'
runtime = '5-00:00:00' # 5 days
memory  = '2096'       # in MB
ncpu    = 1 # nb of threads per job

# parameters of the program:
jdir    = 'job00'
out     = sys.stderr

#-------------------------------------------------------------------------------

def execute(cmd):
    """execute given command with subprocess.call()"""
    try:
        val = subprocess.call(cmd)
        if val:
            out.write("ERROR: command failed with value %i\n" % val)
            print(cmd)
    except OSError:
        out.write("ERROR: command failed:\n")
        print('> '+' '.join(cmd))


def makeNumberedDirectory(root):
    """Create an empty directories that start with 'root'"""
    cnt = 0
    while cnt < 10000:
        name = root + '%02i' % cnt
        try:
            os.mkdir(name)
            return name
        except OSError:
            cnt += 1
    out.write("Error: directory '%s' could not be created \n" % root)
    return ''


def write_script(filename, cmd):
    """create an executable file containing the commands"""
    fid = open(filename, 'w')
    fid.write("#!/bin/bash\n")
    for s in cmd:
        fid.write(s+'\n')
    fid.close()
    os.chmod(filename, 0700)

#------------------------------------------------------------------------

def job(cwd, conf, jarg):
    """return bash script that will run one simulation"""
    cmd  = ['module purge']
    cmd += ['module load foss']
    cmd += ['module load LAPACK OpenBLAS']
    cmd += ['cd %s;' % cwd]
    cmd += ['touch %s;' % conf]
    # the job will call go_sim.py once:
    cmd += ['python go_sim.py %s %s;' % (jarg, conf)]
    cmd += ['mv '+conf+' '+jdir+'/done/.;']
    # cleanup
    cmd += ['mv $0 '+jdir+'/done/.;']
    return cmd


def sub(exe):
    """return command that will submit one job"""
    # specify memory, shell, minimum number of cores and queue
    cmd  = [subcmd, '--nodes=1', '--ntasks=1']
    # specify number of threads if executable is threaded:
    if ncpu > 1:
        cmd += ['--cpus-per-task=%i' % ncpu]
    cmd += ['--job-name='+jdir]
    cmd += ['--partition='+queue] 
    cmd += ['--time='+runtime] 
    cmd += ['--mem='+memory]
    # define signals sent if time is exceeded:
    cmd += ['--signal=15@120']
    cmd += ['--signal=2@60']
    # request special hardware:
    #cmd += ['--constraint=avx2']
    # redirect stderr and sdtout to files:
    cmd += ['--output='+jdir+'/logs/out']
    cmd += ['--error='+jdir+'/logs/err']
    # call executable:
    cmd += [exe]
    return cmd


def array(jobcnt):
    """return command that will submit a job-array"""
    # define parameters directly in the script:
    cmd  = ['#SBATCH --nodes=1']
    cmd += ['#SBATCH --ntasks=1']
    # specify number of threads if executable is threaded:
    if ncpu > 1:
        cmd += ['--cpus-per-task=%i' % ncpu]
    cmd += ['#SBATCH --partition='+queue]
    cmd += ['#SBATCH --time='+runtime]
    cmd += ['#SBATCH --mem='+memory]
    # define signals sent if time is exceeded:
    cmd += ['#SBATCH --signal=INT@60']
    cmd += ['#SBATCH --signal=TERM@120']
    # request special hardware:
    #cmd += ['#SBATCH --constraint=avx2']
    # redirect stderr and sdtout to files:
    cmd += ['#SBATCH --output='+jdir+'/logs/%a.out']
    cmd += ['#SBATCH --error='+jdir+'/logs/%a.err']
    # defines the job name and index range:
    cmd += ['#SBATCH --job-name='+jdir]
    cmd += ['#SBATCH --array=1-%i' % jobcnt]
    cmd += ['']
    cmd += [jdir+'/todo/R$SLURM_ARRAY_TASK_ID']
    return cmd

#-------------------------------------------------------------------------------

def main(args):
    """submit jobs, depending on the arguments provided"""
    global subcmd, memory, runtime, queue, jdir, ncpu
    
    #find subcmd command:
    proc = subprocess.Popen(['which', subcmd], stdout=subprocess.PIPE)
    if proc.wait():
        out.write("Error: submit command `"+subcmd+"' not found!\n")
    else:
        subcmd = proc.stdout.readline().strip()
        #print('|'+subcmd+'|')
    
    # first argument is used for go_sim.py:
    jarg = args.pop(0)
    
    # catch old-style invocation
    if jarg.endswith(".cym"):
        print(__doc__)
        sys.exit()
    
    # make new directories for this job
    jdir = makeNumberedDirectory('job')
    os.mkdir(os.path.join(jdir, 'todo'))
    os.mkdir(os.path.join(jdir, 'done'))
    os.mkdir(os.path.join(jdir, 'save'))
    os.mkdir(os.path.join(jdir, 'logs'))

    print("    go_sim.py will run `%s' in %s" % (jarg, jdir))

    jdup  = 1
    jcnt  = 0
    jname = ''
    cwd   = os.getcwd()
    todo  = os.path.join(jdir,'todo')
    
    for arg in args:
        if os.path.isfile(arg):
            if os.access(arg, os.X_OK):
                out.write("Error: file `%s' should not be executable\n" % arg)
                sys.exit()
            for x in range(jdup):
                jcnt += 1
                conf = todo + '/config%04i.cym' % jcnt
                shutil.copyfile(arg, conf)
                jname = todo + '/R' + str(jcnt)
                cmd = job(cwd, conf, jarg+' name=run%04i'%jcnt+' park='+jdir+'/save')
                write_script(jname, cmd)
        elif arg.isdigit():
            jdup = int(arg)
        elif arg.startswith('mem='):
            memory = arg[4:]
        elif arg.startswith('memory='):
            memory = arg[7:]
        elif arg.startswith('cpu='):
            ncpu = arg[4:]
        elif arg.startswith('ncpu='):
            ncpu = arg[5:]
        elif arg.startswith('day='):
            runtime = arg[4:]+'-00:00:00'
        elif arg.startswith('days='):
            runtime = arg[5:]+'-00:00:00'
        elif arg.startswith('hour='):
            runtime = arg[5:]+':00:00'
        elif arg.startswith('hours='):
            runtime = arg[6:]+':00:00'
        elif arg.startswith('minute='):
            runtime = arg[7:]+':00'
        elif arg.startswith('minutes='):
            runtime = arg[8:]+':00'
        elif arg.startswith('time='):
            runtime = arg[5:]
        elif arg.startswith('queue='):
            queue = arg[6:]
        else:
            out.write("Error: I do not understand argument `%s'\n" % arg)
            sys.exit()

    if memory < 128:
        out.write("Error: requested memory (%s MB) seems too low\n" % memory)
        sys.exit()

    if ncpu < 1:
        out.write("Error: number of cpu/job must be >= 1\n")
        sys.exit()

    if jcnt < 1:
        out.write("Error: you need to specify at least one config file\n" % arg)
        sys.exit()

    if jcnt > 1:
        print("    submit.py created %i scripts in `%s'" % (jcnt, todo))
        cmd = array(jcnt)
        # create script file:
        name = todo + '/job.bash';
        write_script(name, cmd)
        # make command to submit this script:
        cmd = (subcmd, name)
    else:
        cmd = sub(jname);
    execute(cmd)


#-------------------------------------------------------------------------------

if __name__ == "__main__":
    if len(sys.argv) < 2 or sys.argv[1].endswith("help"):
        print(__doc__)
    else:
        main(sys.argv[1:])




#!/usr/bin/env python
#
# reduce.py
#
# copyright F. Nedelec, 2010-2013

"""
reduce.py:
    reduce the size of cytosim file 'objects.cmo' in selected directories
    
Usage:
    reduce.py [min=INTEGER] [commit] DIRECTORIES
    
The word 'commit' must be given otherwise the directories will not be processed
Specifying 'min' sets the minimum number of frames in 'objects.cmo'

Each pass reduces the number of frames by a factor 10.
A file 'reduced' is created in each processed directory.
If such a file already exists, the directory is not processed.
This is a safety mechanism to ensure each directory is processed once only.

F. Nedelec, 28.04.2011 - 20.12.2013
"""

import sys, os, subprocess

commit = 0
min_cnt = 0

err = sys.stderr

#------------------------------------------------------------------------

def nbFrames(file='objects.cmo'):
    try:
        str = subprocess.check_output(['frametool', file])
    except:
        proc = subprocess.Popen(['frametool', file], stdout=subprocess.PIPE)
        line = proc.stdout.readline().split()
        proc.stdout.close()
    try:
        return int(line[0])
    except:
        return int(line[1])
    return 0


def process(path):
    """cleanup directory"""
    os.chdir(path)
    files = os.listdir('.')

    print('- '*32+path)
    if min_cnt > 0:
        cnt = nbFrames()
        if cnt <= min_cnt:
            print(' > objects.cmo has %i frames' % cnt)
            return

    if commit:
        if 'reduced' in files:
            return
        if 'objects.cmo' in files:
            out = open('objectsR.cmo', 'w')
            subprocess.call(['frametool', 'objects.cmo', '0:10:'], stdout=out)
            os.rename('objectsR.cmo', 'objects.cmo');
            print('reduced > objects.cmo')
        if 'messages.cmo' in files:
            out = open('messagesR.cmo', 'w')
            subprocess.call(['grep', '-v', '^F[0-9]*[123456789] ', 'messages.cmo'], stdout=out);
            os.rename('messagesR.cmo', 'messages.cmo');
            print('reduced > messages.cmo')
        subprocess.call(['touch', 'reduced'])
    else:
        if 'objects.cmo' in files:
            cnt = nbFrames('objects.cmo')
            print('  > objects.cmo  : %i frames' % cnt)
        if 'messages.cmo' in files:
            print('  > messages.cmo')


#------------------------------------------------------------------------

def main(args):
    global commit, min_cnt
    paths = []
    
    for arg in args:
        if arg == 'commit':
            commit = 1
        elif arg.startswith('commit='):
            commit = int(arg[7:])
        elif arg.startswith('min='):
            min_cnt = int(arg[4:])
        elif os.path.isdir(arg):
            paths.append(arg)
        else:
            err.write("ignored '%s' on command line\n" % arg)

    if not paths:
        paths.append('.')

    cdir = os.getcwd()
    for path in paths:
        os.chdir(cdir)
        try:
            process(path)
        except Exception as e:
            print(" Error in `%s': %s" % (path, repr(e)));


if __name__ == "__main__":
    if len(sys.argv)>1 and sys.argv[1].endswith("help"):
        print(__doc__)
    else:
        main(sys.argv[1:])



#!/usr/bin/env python
#
# A script to submit analysis jobs to the SLURM queuing system
#
# Derived from submit_slurm.py
# F. Nedelec, 18.11.2018

"""
    Submit a job to the SLURM system to be called in multiple directories
    
Syntax:
    
    submit_one.py ARG [mem=????] [queue=????] [hours=INT] [days=INT] dir1 [dir2] [dir3] [...]
    
    Unless specified otherwise, the queue is 'medium_priority'.
    The amount of requested memory (default=2G) should be specified in MB:
       mem=1024 (for 1 GB)
       mem=512  (for 512 MB)
       ...
    
Example:
    
    submit_one.py 'report platelet > platelet.txt' run????
    Submit one job to run command in the directories provided
    
F. Nedelec, Last updated 18.11.2018
"""


import sys, os, subprocess, tempfile

# default parameters for submission:
subcmd  = 'sbatch'
queue   = 'htc'
runtime = '1:00:00'   # 1 hour
memory  = '4096'      # in MB
ncpu    = 1           # nb of threads per job

# parameters of the program:
out     = sys.stderr

#-------------------------------------------------------------------------------

def execute(cmd):
    """execute given command with subprocess.call()"""
    try:
        val = subprocess.call(cmd)
        if val:
            out.write("ERROR: command failed with value %i\n" % val)
            print(cmd)
    except OSError:
        out.write("ERROR: command failed:\n")
        print('> '+' '.join(cmd))


def write_script(fid, cmd):
    """write that will run the job"""
    os.write(fid, '#!/bin/bash\n')
    os.write(fid, 'module -q purge\n')
    os.write(fid, 'module -q load LAPACK OpenBLAS\n')
    os.write(fid, 'module -q load foss\n')
    for c in cmd:
        os.write(fid, c+'\n')
    os.close(fid)


def sub(file):
    """return command that will submit one job"""
    # specify memory, shell, minimum number of cores and queue
    cmd  = [subcmd, '--nodes=1', '--ntasks=1']
    # specify number of threads if executable is threaded:
    if ncpu > 1:
        cmd += ['--cpus-per-task=%i' % ncpu]
    cmd += ['--partition='+queue]
    cmd += ['--time='+runtime] 
    cmd += ['--mem='+memory]
    # define signals sent if time is exceeded:
    cmd += ['--signal=15@120']
    cmd += ['--signal=2@60']
    # request special hardware:
    cmd += ['--constraint=avx2']
    # redirect stderr and sdtout to files:
    cmd += ['--output='+file+'.out']
    cmd += ['--error='+file+'.err']
    # call script:
    cmd += [file]
    #cmd += ['rm '+file]
    return cmd

#-------------------------------------------------------------------------------

def main(args):
    """submit jobs, depending on the arguments provided"""
    global subcmd, memory, runtime, queue, ncpu
    
    #find subcmd command:
    proc = subprocess.Popen(['which', subcmd], stdout=subprocess.PIPE)
    if proc.wait():
        out.write("Error: submit command `"+subcmd+"' not found!\n")
    else:
        subcmd = proc.stdout.readline().strip()

    # first argument is used for go_sim.py:
    cmd = args.pop(0)

    # create job script file:
    if not os.path.isdir('log'):
        os.mkdir('log')
    fid, file = tempfile.mkstemp('', '', 'log', True)

    job = []
    cwd   = os.getcwd()
    for arg in args:
        if os.path.isdir(arg) and os.access(arg, os.X_OK):
            job += ['cd '+os.path.abspath(arg)+' && '+cmd+';']
        elif arg.startswith('mem='):
            memory = arg[4:]
        elif arg.startswith('memory='):
            memory = arg[7:]
        elif arg.startswith('cpu='):
            ncpu = arg[4:]
        elif arg.startswith('ncpu='):
            ncpu = arg[5:]
        elif arg.startswith('days='):
            runtime = arg[5:]+'-00:00:00'
        elif arg.startswith('hours='):
            runtime = arg[6:]+':00:00'
        elif arg.startswith('minutes='):
            runtime = arg[8:]+':00'
        elif arg.startswith('time='):
            runtime = arg[5:]
        elif arg.startswith('queue='):
            queue = arg[6:]
        else:
            out.write("Error: I do not understand argument `%s'\n" % arg)
            sys.exit()

    if memory < 64:
        out.write("Error: requested memory (%s MB) seems too low\n" % memory)
        sys.exit()

    if ncpu < 1:
        out.write("Error: number of cpu/job must be >= 1\n")
        sys.exit()

    if job:
        write_script(fid, job)
        os.chmod(file, 0700)
        execute(sub(file))
    else:
        out.write("Error: you need to specify at least one directory\n" % arg)


#-------------------------------------------------------------------------------

if __name__ == "__main__":
    if len(sys.argv) < 2 or sys.argv[1].endswith("help"):
        print(__doc__)
    else:
        main(sys.argv[1:])




#!/usr/bin/env python3
#
# make_page.py creates a HTML page with links to files in given directories
#
# Copyright FJ Nedelec, 14.12.2007 -- 4.2015 & 7.3.2022, 15.3.2023

"""
Synopsis:

    Generates an HTML page linking images & movies found in sub-directories

Usage:
    
    make_page.py [recursive=1] [tile=INT] [width=INT] [height=INT] [OUTPUT] DIRECTORIES/FILES

Info:
    
    If `tile` is set, a HTML table will be generated, and the value specified will be the
    number of entries in each line. The name of the ouput file ('page.html' by default), 
    can be specified on the command line (extension should be .html).
    `width` or `height` specify the size in pixels at which images will appear on the HTML.


Copyright F.J. Nedelec, Cambridge University
Created  14.12.2007
Modified 3.2010, 5.2012, 11.2012, 7.2013, 11.2013, 4.2015, 7.3.2022, 11.3.2023
"""

import sys, os, subprocess

out  = 0
indx = 1
iarg = ''
tile = 0
curs = 0
excluded = []
master = 'config.cym'
output = 'page.html'
back_color = '#111111'
text_color = '#AAAAAA'

def writeHeader():
    global out;
    out.write('<html>\n')
    out.write('<head>\n')
    try:
        out.write('<title>%s</title>\n'%os.getcwd())
    except Exception:
        pass
    out.write('<meta name="ROBOTS" content="NOINDEX, NOFOLLOW">\n')
    out.write('<script type="text/JavaScript">\n')
    out.write('var zwin=0;\n')
    out.write('function mk_win()   { zwin=window.open("","see","resizable=1,width=1280,height=1024"); }\n')
    #out.write('function set(i, m)  { document.images[i].src=m; }\n')
    out.write('function zoom(m)    { mk_win(); zwin.location=m; }\n')
    #out.write('function stat(s)    { window.status=s; }\n')
    out.write('</script>\n')
    out.write('</head>\n')
    out.write(f'<body bgcolor="{back_color}">\n')
    out.write('<h2>\n')
    out.write('<a href="index.html">index</a>\n')
    out.write('</h2>\n')


def writeFooter():
    global out;
    out.write('</body>\n')
    out.write('</html>\n')


def getMovieSize(file):
    """
    Call ffprobe, which is a tool that comes with ffmpeg,
    and parse output to extract size of videos or images
    """
    W = 256
    H = 256
    call = subprocess.run(['ffprobe', '-v', 'quiet', '-show_streams', file], capture_output=True)
    for line in call.stdout.decode():
        #print(line, end='')
        try:
            [key, equal, val] = line.partition('=')
            if key=="width":
                W = int(val)
            elif key=="height":
                H = int(val)
        except:
            pass
    #print(" `%s' is %i x %i\n" %(file, W, H))
    return W, H


def writeImageLinks(files):
    global out, iarg
    for f in sorted(files):
        shot = os.path.basename(f)
        #out.write('<br>%s: ' % shot)
        out.write('<a href="javascript:zoom(\'%s\');">\n' % f)
        out.write('  <img %s src="%s" alt="%s">\n' % (iarg, f, shot))
        out.write('</a>\n')
    if files:
        out.write('\n')


def writeMovieLinks(files):
    global out
    for f in sorted(files):
        W, H = getMovieSize(f)
        shot = os.path.basename(f)
        out.write('<video controls="controls" width="%s" height="%s" loop="true" alt="%s">\n' % (W, H, shot))
        out.write('  <source src="%s" type="video/mp4">\n' % f)
        out.write('</video>\n')
    if files:
        out.write('\n')


def writeDifferences(left, right):
    """
    Extract differences in 'right' compare to left, and print them to 'out'
    """
    if not os.path.isfile(left) or not os.path.isfile(right):
        return
    sub = subprocess.Popen(['diff', left, right], stdout=subprocess.PIPE)
    out.write(f'<pre style="color:{text_color}">\n')
    for data in sub.stdout:
        try:
            line = data.decode('utf-8')
        except:
            line = data
        if line[0] == '>' and not line[1:].isspace():
            out.write(line)
    sub.stdout.close()
    out.write('</pre>\n')


def process(dirpath, subdir, files, images, movies):
    """
    Write HTML code for given directory
    """
    global out, indx, tile, curs
    if tile > 0:
        out.write('<td>\n')
    out.write('<h3 style="color:Yellow;padding:3px;margin:3px">\n')
    out.write(dirpath.lstrip('./'))
    for f in files:
        out.write(' &mdash; <a href="%s/%s">%s</a>' % (dirpath, f, f))
    out.write('\n</h3>\n')
    if master in files:
        writeDifferences(master, os.path.join(dirpath, master))
    writeImageLinks(images)
    writeMovieLinks(movies)
    if curs:
        for d in subdir:
            process_dir(d)
    if tile > 0:
        out.write('</td>\n')
        if indx % tile == 0:
            out.write('</tr><tr>\n')
    indx += 1


def process_dir(dirpath):
    """select relevant files in given sub-directory and call process()"""
    subdir = []
    files = []
    images = []
    movies = []
    for f in os.listdir(dirpath):
        path = os.path.join(dirpath, f)
        if os.path.isdir(path):
            subdir.append(f)
        else:
            [name, ext] = os.path.splitext(f)
            ext = ext.lower()
            if name in excluded:
                pass
            elif ext in ['.png', '.jpg', '.gif', '.tif', '.svg']:
                images.append(path)
            elif ext in ['.mp4', '.mov']:
                movies.append(path)
            elif f in [master, 'movie.mp4']:
                files.append(f)
    process(dirpath, subdir, files, images, movies)


#------------------------------------------------------------------------

def main(args):
    """generates HTML page"""
    global output, out, iarg, tile, curs, master
    paths = []
    
    for arg in args:
        [key, equal, val] = arg.partition('=')
        if key and equal=='=' and val:
            if key=='width' or key=='height':
                iarg = arg
            elif key=='table':
                tile = int(val)
            elif key=='tile':
                tile = int(val)
            elif key=='recursive':
                curs = int(val)
            elif key=='exclude':
                excluded.append(val)
            elif key=='master':
                master = val
            else:
                sys.stderr.write("ignored '%s' on command line\n" % arg)
        else:
            if os.path.isdir(arg):
                paths.append(arg)
            elif arg.endswith('.html'):
                output = arg
            else:
                sys.stderr.write("ignored '%s' on command line\n" % arg)

    if not paths:
        sys.stderr.write("You must specify a path on the command line\n")
        sys.exit()

    try:
        out = open(output, 'w')
    except Exception as e:
        sys.stderr.write("Error creating file `%s': %s\n" % (output, repr(e)))
        out = sys.stdout
    writeHeader()
    if tile > 0:
        out.write('<table border="0" align="center" cellpadding="1">\n')
        out.write('<tr>\n')
    for p in paths:
        process_dir(p)
    if tile > 0:
        out.write('</tr>\n')
        out.write('</table>\n')
    writeFooter()
    if out != sys.stdout:
        out.close()
        print("generated '%s' with %i entries" % (output, indx-1))


#------------------------------------------------------------------------

if __name__ == "__main__":
    if len(sys.argv)>1 and sys.argv[1].endswith("help"):
        print(__doc__)
    else:
        main(sys.argv[1:])




#!/usr/bin/env python3
#
# compare.py
#
# copyright F. Nedelec, 14.12.2007, 14.03.2018; 4.8.2020, 18.10.2023

"""
compare.py
    Compare source files contained in two root directories
Usage:
    compare.py root1 root2 [opendiff]
"""

import sys, os, shutil, time, tempfile
import difflib, subprocess

mode="diff"
diff="diff --side-by-side -W200 -p --suppress-common-lines"



def print_spacer(arg):
    """print a line of width size, with 'arg' in the middle"""
    rows, cols = os.popen('stty size', 'r').read().split()
    sys.stdout.write(chr(27)+"[36;2m"); sys.stdout.flush()
    print(arg.center(int(cols), '-'))
    sys.stdout.write(chr(27)+"[0m"); sys.stdout.flush()


def compareFiles(fileL, fileR):
    if not os.path.isfile(fileL):
        return print("missing < %s" % fileL)
    if not os.path.isfile(fileR):
        return print("missing > %s" % fileR)
    #print("compare %s %s" % (fileL, fileR))
    comp = os.popen(diff+" "+fileL+" "+fileR)
    difLR = comp.read()
    comp.close()
    if difLR:
        if mode == 'diff':
            print_spacer(os.path.basename(fileL))
            for line in difLR:
                print(line, end='')
            
            sys.stdout.write(chr(27)+"[33;1m")
            print("This was `%s`" % os.path.basename(fileL), end=':')
            ans = input(' (return/left/right/open/q) ?>'+chr(27)+'[0m')
            
            if ans == "left" or ans == "l":
                shutil.copyfile(fileL, fileR)
            elif ans == "right" or ans == "r":
                shutil.copyfile(fileR, fileL)
            elif ans == "swap":
                fid, tmp = tempfile.mkstemp('.txt', 'temp', '', True)
                print(os.getcwd(), os.path.isfile(file))
                os.close(fid);
                os.rename(fileL, tmp)
                os.rename(fileR, fileL)
                os.rename(tmp, fileR)
            elif ans == "open":
                os.system("opendiff "+fileL+" "+fileR+"&")
            elif ans == "q":
                sys.exit()
        else:
            subprocess.call(["opendiff", fileL, fileR])
            #we wait a bit for the application to start
            time.sleep(0.5)


def interesting(file):
    return ( file.endswith('.py') or file.endswith('.m')
        or file.endswith('.h') or file.endswith('.cc')
        or file.endswith('.md') or file.endswith('.txt')
        or file.startswith('makefile') or file.startswith('.cym') )


def process_dir(roots, pathL, files):
    """compare files in the current directory"""
    if pathL.startswith(roots[0]):
        path = pathL[len(roots[0]):]
    else:
        path = pathL
    pathR = os.path.normpath(roots[1] + '/' + path)
    #print("PATH %s %s %s" % (path, pathL, pathR))
    #print("FILES %s" % files)
    if path.endswith('.svn'):
        return
    if path.endswith('.git'):
        return
    if 'DerivedData/' in path:
        return
    if 0 <= path.find('/.git/'):
        return
    if path.startswith('/bin'):
        return
    if path.startswith('/build/'):
        return
    print_spacer('%s %s'%(pathL, pathR))
    for file in files:
        if interesting(file):
            fileL = os.path.join(pathL, file)
            fileR = os.path.join(pathR, file)
            compareFiles(fileL, fileR)

#------------------------------------------------------------------------------

def main(args):
    """main"""
    global mode
    if len(args) < 2:
        print("Error: you must specify two root directories!")
        sys.exit()
    rootL = os.path.relpath(args[0])
    rootR = os.path.relpath(args[1])
    if not os.path.isdir(rootL):
        print("Error: `%s' is not a directory" % rootL)
        sys.exit()
    if not os.path.isdir(rootR):
        print("Error: `%s' is not a directory" % rootR)
        sys.exit()
    #parse command-line arguments:    
    for arg in args[2:]:
        if arg == 'opendiff':
            mode=arg
        else:
            print("unknown argument '%s'" % arg)
            sys.exit()
    # process all directories within 'root':
    #print("Comparing %s and %s" % (rootL, rootR))
    for path, dirs, files in os.walk(rootL, topdown=False):
        process_dir([rootL, rootR], path, files)



if __name__ == "__main__":
    if len(sys.argv) < 3 or sys.argv[1]=='help':
        print(__doc__)
    else:
        main(sys.argv[1:])

